# latent-feature based collaborative filtering
# copyright David Poole 2012
# Released under GPL
# Based on gradient descent algorithm of Koren, Y., Bell, R. and Volinsky, C., 
#   Matrix Factorization Techniques for Recommender Systems, IEEE Computer 2009

# Data set is from movielense (http://www.grouplens.org/node/73). Uncomment one of the following two lines.
from select_ratings import used_ratings as ratings # useful for plotting all ratings
#from movielens_100k_data import ratings # full data set
import random
import matplotlib.pyplot as plt

step_size = 0.01     # gradient descent step size
num_iter = 50        # number of iterations of gradient descent
regularization = 0.1 # the weight for the regulization terms
num_features = 10    # number of hidden features
feature_range = 0.2  # feaures are initialized to be between -feature_range and feature_range

num_ratings = len(ratings)
ave_rating = (sum([rating for (user,item,rating,timestamp) in ratings])*1.0)/num_ratings

set_users_that_rated = set([user for (user,item,rating,timestamp) in ratings]) # users who rated
set_items_rated = set([item for (user,item,rating,timestamp) in ratings]) # items that were rated
max_item_index = max(set_items_rated)
max_user_index = max(set_users_that_rated)

user_offset = [0]*(max_user_index+1)
item_offset = [0]*(max_item_index+1)
user_feature = [[random.uniform(-feature_range,feature_range) for f in range(num_features)] for i in range(max_user_index+1)]
item_feature = [[random.uniform(-feature_range,feature_range) for f in range(num_features)] for i in range(max_item_index+1)]
iter=0

def doit(num_it):
  """ do num_it iterations of gradient descent."""
  global iter
  for i in range(num_it):
    iter += 1
    total_error=0
    sumsq_error=0
    for (user,item,rating,timestamp) in ratings:
        prediction = ave_rating + user_offset[user] + item_offset[item] + sum([user_feature[user][f]*item_feature[item][f] for f in range(num_features)])
        error = prediction - rating
        total_error += abs(error)
        sumsq_error += error * error
        user_offset[user] += -step_size*error - step_size*regularization* user_offset[user]
        item_offset[item] += -step_size*error - step_size*regularization*item_offset[item]
        for f in range(num_features):
            user_f = user_feature[user][f]
            user_feature[user][f] += -step_size*error*item_feature[item][f] -step_size*regularization* user_feature[user][f]
            item_feature[item][f] += -step_size*error*user_f -step_size*regularization*item_feature[item][f]
    print "Iteration",iter,"Ave Error =",total_error/num_ratings,"Ave sq error =",sumsq_error/num_ratings

doit(num_iter)

def plot_feature(f,num_points=200):
    """plot some of the user-movie ratings, 
    num_points is the number of points selected at random plotted.

    the plot has the users on the x-axis sorted by their value on feature f and
    with the items on the y-axis sorted by their value on feature f and 
    the ratings plotted at the corresponding x-y position.
    """
    user_vals = [user_feature[user][f] for user in range(max_user_index+1) if user in set_users_that_rated]
    item_vals = [item_feature[item][f] for item in range(max_item_index+1) if item in set_items_rated]
    plt.axis([min(user_vals)-0.1,max(user_vals)+0.1,min(item_vals)-0.1,max(item_vals)+0.1])
    for i in range(num_points):
        (user,item,rating,timestamp) = random.choice(ratings)
        plt.text(user_feature[user][f], item_feature[item][f], str(rating))
    plt.show()

def plot_all_feature(f):
    """plot all of the user-movie ratings, 

    the plot has the users on the x-axis sorted by their value on feature f and
    with the items on the y-axis sorted by their value on feature f and 
    the ratings plotted at the corresponding x-y position.
    """
    user_vals = [user_feature[user][f] for user in range(max_user_index+1) if user in set_users_that_rated]
    item_vals = [item_feature[item][f] for item in range(max_item_index+1) if item in set_items_rated]
    plt.axis([min(user_vals)-0.1,max(user_vals)+0.1,min(item_vals)-0.1,max(item_vals)+0.1])
    for (user,item,rating,timestamp) in ratings:
        plt.text(user_feature[user][f], item_feature[item][f], str(rating))
    plt.show()

#learned_user_features = [user_feature[user][0] for user in range(max_user_index+1) if user in set_users_that_rated]
#learned_item_features = [item_feature[item][0] for item in range(max_item_index+1) if item in set_items_rated]
learned_user_offsets = [user_offset[user] for user in range(max_user_index+1) if user in set_users_that_rated]
learned_item_offsets = [item_offset[item] for item in range(max_item_index+1) if item in set_items_rated]

#>>> user_offsets_1f = [user_offset[user] for user in range(max_user_index+1) if user in set_users_that_rated]
#>>> item_offsets_1f = [item_offset[item] for item in range(max_item_index+1) if item in set_items_rated]
# [x-y for (x,y) in zip(learned_item_offsets,item_offsets_1f)]

# try
# plot_all_feature(0)
